#
#   JSON parser for Tcl (http://www.tcl.tk/software/tcllib/)
#
#   See http://www.json.org/ && http://www.ietf.org/rfc/rfc4627.txt
#
#   Total rework of the code published with version number 1.0 by
#   Thomas Maeder, Glue Software Engineering AG
#   Included here (and stripped down to just json2dict) by Anthony Mallet to
#   avoid dealing with packaging issues on various unix distributions.
#
#   $Id: json.tcl,v 1.7 2011/11/10 21:05:58 andreas_kupries Exp $
#
# This software is copyrighted by Ajuba Solutions and other parties.
#
# The authors hereby grant permission to use, copy, modify, distribute,
# and license this software and its documentation for any purpose, provided
# that existing copyright notices are retained in all copies and that this
# notice is included verbatim in any distributions. No written agreement,
# license, or royalty fee is required for any of the authorized uses.
# Modifications to this software may be copyrighted by their authors
# and need not follow the licensing terms described here, provided that
# the new terms are clearly indicated on the first page of each file where
# they apply.
#
# IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
# DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
# IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.
#
# GOVERNMENT USE: If you are acquiring this software on behalf of the
# U.S. government, the Government shall have only "Restricted Rights"
# in the software and related documentation as defined in the Federal
# Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
# are acquiring the software on behalf of the Department of Defense, the
# software shall be classified as "Commercial Computer Software" and the
# Government shall have only "Restricted Rights" as defined in Clause
# 252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
# authors grant the U.S. Government and others acting in its behalf
# permission to use and distribute the software in accordance with the
# terms specified in this license.

namespace eval genomix::json {
  # Regular expression for tokenizing a JSON text (cf. http://json.org/)

  # tokens consisting of a single character
  variable singleCharTokens { "{" "}" ":" "\\[" "\\]" "," }
  variable singleCharTokenRE "\[[join $singleCharTokens {}]\]"

  # quoted string tokens
  variable escapableREs { "[\\\"\\\\/bfnrt]" "u[[:xdigit:]]{4}" }
  variable escapedCharRE "\\\\(?:[join $escapableREs |])"
  variable unescapedCharRE {[^\\\"]}
  variable stringRE "\"(?:$escapedCharRE|$unescapedCharRE)*\""

  # (unquoted) words
  variable wordTokens { "true" "false" "null" }
  variable wordTokenRE [join $wordTokens "|"]

  # number tokens
  # negative lookahead (?!0)[[:digit:]]+ might be more elegant, but
  # would slow down tokenizing by a factor of up to 3!
  variable positiveRE {[1-9][[:digit:]]*}
  variable cardinalRE "-?(?:$positiveRE|0)"
  variable fractionRE {[.][[:digit:]]+}
  variable exponentialRE {[eE][+-]?[[:digit:]]+}
  variable numberRE "${cardinalRE}(?:$fractionRE)?(?:$exponentialRE)?"

  # JSON token
  variable tokenRE "$singleCharTokenRE|$stringRE|$wordTokenRE|$numberRE"

  # 0..n white space characters
  set whiteSpaceRE {[[:space:]]*}

  # Regular expression for validating a JSON text
  variable validJsonRE "^(?:${whiteSpaceRE}(?:$tokenRE))*${whiteSpaceRE}$"


  # Parse JSON text into a dict
  # @param jsonText JSON text
  # @return dict (or list) containing the object represented by $jsonText
  proc json2dict {jsonText} {
    variable tokenRE

    set tokens [regexp -all -inline -- $tokenRE $jsonText]
    set nrTokens [llength $tokens]
    set tokenCursor 0
    return [parseValue $tokens $nrTokens tokenCursor]
  }

  # Throw an exception signaling an unexpected token
  proc unexpected {tokenCursor token expected} {
    return -code error "unexpected token \"$token\" at position $tokenCursor; expecting $expected"
  }

  # Get rid of the quotes surrounding a string token and substitute the
  # real characters for escape sequences within it
  # @param token
  # @return unquoted unescaped value of the string contained in $token
  proc unquoteUnescapeString {token} {
    set unquoted [string range $token 1 end-1]
    return [subst -nocommands -novariables $unquoted]
  }

  # Parse an object member
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @param objectDictName name (in caller's context) of dict
  #                       representing the JSON object of which to
  #                       parse the next member
  proc parseObjectMember {tokens nrTokens tokenCursorName objectDictName} {
    upvar $tokenCursorName tokenCursor
    upvar $objectDictName objectDict

    set token [lindex $tokens $tokenCursor]
    incr tokenCursor

    set leadingChar [string index $token 0]
    if {$leadingChar eq "\""} {
      set memberName [unquoteUnescapeString $token]

      if {$tokenCursor == $nrTokens} {
        unexpected $tokenCursor "END" "\":\""
      } else {
        set token [lindex $tokens $tokenCursor]
        incr tokenCursor

        if {$token eq ":"} {
          set memberValue [parseValue $tokens $nrTokens tokenCursor]
          dict set objectDict $memberName $memberValue
        } else {
          unexpected $tokenCursor $token "\":\""
        }
      }
    } else {
      unexpected $tokenCursor $token "STRING"
    }
  }

  # Parse the members of an object
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @param objectDictName name (in caller's context) of dict
  #                       representing the JSON object of which to
  #                       parse the next member
  proc parseObjectMembers {tokens nrTokens tokenCursorName objectDictName} {
    upvar $tokenCursorName tokenCursor
    upvar $objectDictName objectDict

    while true {
      parseObjectMember $tokens $nrTokens tokenCursor objectDict

      set token [lindex $tokens $tokenCursor]
      incr tokenCursor

      switch -exact $token {
        "," {
          # continue
        }
        "\}" {
          break
        }
        default {
          unexpected $tokenCursor $token "\",\"|\"\}\""
        }
      }
    }
  }

  # Parse an object
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @return parsed object (Tcl dict)
  proc parseObject {tokens nrTokens tokenCursorName} {
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
      unexpected $tokenCursor "END" "OBJECT"
    } else {
      set result [dict create]

      set token [lindex $tokens $tokenCursor]

      if {$token eq "\}"} {
        # empty object
        incr tokenCursor
      } else {
        parseObjectMembers $tokens $nrTokens tokenCursor result
      }

      return $result
    }
  }

  # Parse the elements of an array
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @param resultName name (in caller's context) of the list
  #                   representing the JSON array
  proc parseArrayElements {tokens nrTokens tokenCursorName resultName} {
    upvar $tokenCursorName tokenCursor
    upvar $resultName result

    while true {
      lappend result [parseValue $tokens $nrTokens tokenCursor]

      if {$tokenCursor == $nrTokens} {
        unexpected $tokenCursor "END" "\",\"|\"\]\""
      } else {
        set token [lindex $tokens $tokenCursor]
        incr tokenCursor

        switch -exact $token {
          "," {
            # continue
          }
          "\]" {
            break
          }
          default {
            unexpected $tokenCursor $token "\",\"|\"\]\""
          }
        }
      }
    }
  }

  # Parse an array
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @return parsed array (Tcl list)
  proc parseArray {tokens nrTokens tokenCursorName} {
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
      unexpected $tokenCursor "END" "ARRAY"
    } else {
      set result {}

      set token [lindex $tokens $tokenCursor]

      set leadingChar [string index $token 0]
      if {$leadingChar eq "\]"} {
        # empty array
        incr tokenCursor
      } else {
        parseArrayElements $tokens $nrTokens tokenCursor result
      }

      return $result
    }
  }

  # Parse a value
  # @param tokens list of tokens
  # @param nrTokens length of $tokens
  # @param tokenCursorName name (in caller's context) of variable
  #                        holding current position in $tokens
  # @return parsed value (dict, list, string, number)
  proc parseValue {tokens nrTokens tokenCursorName} {
    upvar $tokenCursorName tokenCursor

    if {$tokenCursor == $nrTokens} {
      unexpected $tokenCursor "END" "VALUE"
    } else {
      set token [lindex $tokens $tokenCursor]
      incr tokenCursor

      set leadingChar [string index $token 0]
      switch -exact -- $leadingChar {
        "\{" {
          return [parseObject $tokens $nrTokens tokenCursor]
        }
        "\[" {
          return [parseArray $tokens $nrTokens tokenCursor]
        }
        "\"" {
          # quoted string
          return [unquoteUnescapeString $token]
        }
        "t" -
        "f" -
        "n" {
          # bare word: true, false or null
          return $token
        }
        default {
          # number?
          if {[string is double -strict $token]} {
            return $token
          } else {
            unexpected $tokenCursor $token "VALUE"
          }
        }
      }
    }
  }
}
